@@ -1,5 +1,4 @@
 /* Links viewing/manipulation handling */
-/* $Id: link.c,v 1.309.2.6 2005/05/01 21:00:42 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -22,9 +21,10 @@
 #include "document/view.h"
 #include "ecmascript/ecmascript.h"
 #include "intl/gettext/libintl.h"
+#include "main/object.h"
 #include "protocol/uri.h"
-#include "sched/session.h"
-#include "sched/task.h"
+#include "session/session.h"
+#include "session/task.h"
 #include "terminal/color.h"
 #include "terminal/draw.h"
 #include "terminal/kbd.h"
@@ -34,8 +34,8 @@
 #include "util/conv.h"
 #include "util/error.h"
 #include "util/memory.h"
-#include "util/object.h"
 #include "util/string.h"
+#include "viewer/action.h"
 #include "viewer/text/form.h"
 #include "viewer/text/link.h"
 #include "viewer/text/search.h"
@@ -157,34 +157,27 @@ init_link_drawing(struct document_view *doc_view, struct link *link, int invert)
 
 	template->attr = SCREEN_ATTR_STANDOUT;
 
-	/* For the color mode options we use the options set for the document.
-	 * But for the active link options we prefer to use the global
-	 * global_doc_opts since it is kept up to date by an option change
-	 * hook. However if it is not available fall back to use the options
-	 * from the viewed document. */
 	doc_opts = &doc_view->document->options;
 
 	color_flags = (doc_opts->color_flags | COLOR_DECREASE_LIGHTNESS);
 	color_mode = doc_opts->color_mode;
 
-	if (global_doc_opts) doc_opts = global_doc_opts;
-
-	if (doc_opts->underline_active_link)
+	if (doc_opts->active_link.underline)
 		template->attr |= SCREEN_ATTR_UNDERLINE;
 
-	if (doc_opts->bold_active_link)
+	if (doc_opts->active_link.bold)
 		template->attr |= SCREEN_ATTR_BOLD;
 
-	if (doc_opts->color_active_link) {
-		colors.foreground = doc_opts->active_link_fg;
-		colors.background = doc_opts->active_link_bg;
+	if (doc_opts->active_link.color) {
+		colors.foreground = doc_opts->active_link.fg;
+		colors.background = doc_opts->active_link.bg;
 	} else {
 		colors.foreground = link->color.foreground;
 		colors.background = link->color.background;
 	}
 
-	if (doc_opts->invert_active_link && invert) {
-		swap_values(color_t, colors.foreground, colors.background);
+	if (invert && doc_opts->active_link.invert) {
+		swap_values(color_T, colors.foreground, colors.background);
 
 		/* Highlight text-input form-fields correctly if contrast
 		 * correction is needed. */
@@ -376,7 +369,7 @@ get_last_link(struct document_view *doc_view)
 	return link;
 }
 
-static inline int
+static int
 link_in_view_x(struct document_view *doc_view, struct link *link)
 {
 	int i, dx, width;
@@ -397,7 +390,7 @@ link_in_view_x(struct document_view *doc_view, struct link *link)
 	return 0;
 }
 
-int
+static int
 link_in_view_y(struct document_view *doc_view, struct link *link)
 {
 	int i, dy, height;
@@ -418,7 +411,7 @@ link_in_view_y(struct document_view *doc_view, struct link *link)
 	return 0;
 }
 
-int
+static int
 link_in_view(struct document_view *doc_view, struct link *link)
 {
 	assert(doc_view && link);
@@ -462,10 +455,10 @@ get_visible_links_range(struct document_view *doc_view, int *first, int *last)
 	}
 }
 
-int
-next_link_in_view(struct document_view *doc_view, int current, int direction,
-	          int (*fn)(struct document_view *, struct link *),
-	          void (*cntr)(struct document_view *, struct link *))
+static int
+next_link_in_view_(struct document_view *doc_view, int current, int direction,
+	           int (*fn)(struct document_view *, struct link *),
+	           void (*cntr)(struct document_view *, struct link *))
 {
 	struct document *document;
 	struct view_state *vs;
@@ -497,6 +490,18 @@ next_link_in_view(struct document_view *doc_view, int current, int direction,
 	return 0;
 }
 
+int
+next_link_in_view(struct document_view *doc_view, int current, int direction)
+{
+	return next_link_in_view_(doc_view, current, direction, link_in_view, NULL);
+}
+
+int
+next_link_in_view_y(struct document_view *doc_view, int current, int direction)
+{
+	return next_link_in_view_(doc_view, current, direction, link_in_view_y, set_pos_x);
+}
+
 /* Get the bounding columns of @link at line @y (or all lines if @y == -1). */
 static void
 get_link_x_bounds(struct link *link, int y, int *min_x, int *max_x)
@@ -788,7 +793,7 @@ find_link(struct document_view *doc_view, int direction, int page_mode)
 	link_pos = link - doc_view->document->links;
 	if (page_mode) {
 		/* PAGE */
-		next_link_in_view(doc_view, link_pos, direction, link_in_view, NULL);
+		next_link_in_view(doc_view, link_pos, direction);
 		return;
 	}
 	current_link_blur(doc_view);
@@ -883,10 +888,83 @@ goto_current_link(struct session *ses, struct document_view *doc_view, int do_re
 	return link;
 }
 
+static enum frame_event_status
+activate_link(struct session *ses, struct document_view *doc_view,
+              struct link *link, int do_reload)
+{
+	struct form_control *link_fc;
+	struct form_state *fs;
+	struct form *form;
+
+	switch (link->type) {
+	case LINK_HYPERTEXT:
+	case LINK_MAP:
+	case LINK_FIELD:
+	case LINK_AREA:
+	case LINK_BUTTON:
+		if (goto_current_link(ses, doc_view, do_reload))
+			return FRAME_EVENT_OK;
+
+		break;
+	case LINK_CHECKBOX:
+		link_fc = get_link_form_control(link);
+
+		if (form_field_is_readonly(link_fc))
+			return FRAME_EVENT_OK;
+
+		fs = find_form_state(doc_view, link_fc);
+		if (!fs) return FRAME_EVENT_OK;
+
+		if (link_fc->type == FC_CHECKBOX) {
+			fs->state = !fs->state;
+
+			return FRAME_EVENT_REFRESH;
+		}
+
+		foreach (form, doc_view->document->forms) {
+			struct form_control *fc;
+
+			if (form != link_fc->form)
+				continue;
+
+			foreach (fc, form->items) {
+				if (fc->type == FC_RADIO
+				    && !xstrcmp(fc->name, link_fc->name)) {
+					struct form_state *frm_st;
+
+					frm_st = find_form_state(doc_view, fc);
+					if (frm_st) frm_st->state = 0;
+				}
+			}
+		}
+		fs->state = 1;
+
+		break;
+
+	case LINK_SELECT:
+		link_fc = get_link_form_control(link);
+
+		if (form_field_is_readonly(link_fc))
+			return FRAME_EVENT_OK;
+
+		object_lock(doc_view->document);
+		add_empty_window(ses->tab->term,
+				 (void (*)(void *)) release_document,
+				 doc_view->document);
+		do_select_submenu(ses->tab->term, link_fc->menu, ses);
+
+		break;
+
+	default:
+		INTERNAL("bad link type %d", link->type);
+	}
+
+	return FRAME_EVENT_REFRESH;
+}
+
 enum frame_event_status
 enter(struct session *ses, struct document_view *doc_view, int do_reload)
 {
-	struct form_control *link_fc;
 	struct link *link;
 
 	assert(ses && doc_view && doc_view->vs && doc_view->document);
@@ -895,66 +973,10 @@ enter(struct session *ses, struct document_view *doc_view, int do_reload)
 	link = get_current_link(doc_view);
 	if (!link) return FRAME_EVENT_REFRESH;
 
+
 	if (!current_link_evhook(doc_view, SEVHOOK_ONCLICK))
 		return FRAME_EVENT_REFRESH;
-
-	if (!link_is_form(link)
-	    || link_is_textinput(link)
-	    || link->type == LINK_BUTTON) {
-		if (goto_current_link(ses, doc_view, do_reload))
-			return FRAME_EVENT_OK;
-
-		return FRAME_EVENT_REFRESH;
-	}
-
-	link_fc = get_link_form_control(link);
-
-	if (form_field_is_readonly(link_fc))
-		return FRAME_EVENT_OK;
-
-	if (link->type == LINK_CHECKBOX) {
-		struct form_state *fs;
-
-		fs = find_form_state(doc_view, link_fc);
-		if (!fs) return FRAME_EVENT_OK;
-
-		if (link_fc->type == FC_CHECKBOX) {
-			fs->state = !fs->state;
-
-		} else {
-			struct form *form;
-
-			foreach (form, doc_view->document->forms) {
-				struct form_control *fc;
-
-				if (form != link_fc->form)
-					continue;
-
-				foreach (fc, form->items) {
-					if (fc->type == FC_RADIO
-				            && !xstrcmp(fc->name, link_fc->name)) {
-						struct form_state *frm_st;
-
-						frm_st = find_form_state(doc_view, fc);
-						if (frm_st) frm_st->state = 0;
-					}
-				}
-			}
-			fs->state = 1;
-		}
-
-	} else if (link->type == LINK_SELECT) {
-		object_lock(doc_view->document);
-		add_empty_window(ses->tab->term,
-				 (void (*)(void *)) release_document,
-				 doc_view->document);
-		do_select_submenu(ses->tab->term, link_fc->menu, ses);
-
-	} else {
-		INTERNAL("bad link type %d", link->type);
-	}
-
-	return FRAME_EVENT_REFRESH;
+	return activate_link(ses, doc_view, link, do_reload);
 }
 
 struct link *
@@ -1077,19 +1099,19 @@ try_document_key(struct session *ses, struct document_view *doc_view,
 	assert(ses && doc_view && doc_view->document && doc_view->vs && ev);
 	if_assert_failed return FRAME_EVENT_IGNORED;
 
-	if (isasciialpha(get_kbd_key(ev)) && !check_kbd_modifier(ev, KBD_ALT)) {
+	if (!check_kbd_modifier(ev, KBD_MOD_ALT)) {
 		/* We accept those only in alt-combo. */
 		return FRAME_EVENT_IGNORED;
 	}
 
 	/* Run through all the links and see if one of them is bound to the
 	 * key we test.. */
-	key = toupper(get_kbd_key(ev));
+	key = get_kbd_key(ev);
 
 	for (i = 0; i < doc_view->document->nlinks; i++) {
 		struct link *link = &doc_view->document->links[i];
 
-		if (key == link->accesskey) {
+		if (key == link->accesskey) {	/* FIXME: key vs unicode ... */
 			if (passed != i && i <= doc_view->vs->current_link) {
 				/* This is here in order to rotate between
 				 * links with same accesskey. */
@@ -1178,9 +1200,10 @@ link_menu(struct terminal *term, void *xxx, void *ses_)
 				struct string keystroke;
 
 				if (init_string(&keystroke))
-					add_keystroke_to_string(&keystroke,
-								ACT_EDIT_OPEN_EXTERNAL,
-								KEYMAP_EDIT);
+					add_keystroke_action_to_string(
+					                 &keystroke,
+					                 ACT_EDIT_OPEN_EXTERNAL,
+							 KEYMAP_EDIT);
 
 				add_to_menu(&mi, N_("Open in ~external editor"),
 					    keystroke.source, ACT_MAIN_NONE,
@@ -1207,6 +1230,9 @@ link_menu(struct terminal *term, void *xxx, void *ses_)
 
 			add_menu_action(&mi, N_("~Reset form"), ACT_MAIN_RESET_FORM);
 		}
+
+		add_to_menu(&mi, N_("Form f~ields"), NULL, ACT_MAIN_LINK_FORM_MENU,
+			    NULL, NULL, SUBMENU);
 	}
 
 	if (link->where_img) {
@@ -1260,11 +1286,7 @@ get_current_link_info(struct session *ses, struct document_view *doc_view)
 
 	/* TODO: Provide info about script event hooks too. --pasky */
 
-	if (link_is_form(link)) {
-		if (!get_link_form_control(link)) return NULL;
-
-		return get_form_info(ses, doc_view);
-	} else {
+	if (!link_is_form(link)) {
 		struct terminal *term = ses->tab->term;
 		struct string str;
 		unsigned char *uristring = link->where;
@@ -1286,11 +1308,15 @@ get_current_link_info(struct session *ses, struct document_view *doc_view)
 		if (link->accesskey > 0
 		    && get_opt_bool("document.browse.accesskey.display")) {
 			add_to_string(&str, " (");
-			make_keystroke(&str, link->accesskey, 0, 0);
+			add_accesskey_to_string(&str, link->accesskey);
 			add_char_to_string(&str, ')');
 		}
 
 		decode_uri_string_for_display(&str);
 		return str.source;
 	}
+
+	if (!get_link_form_control(link)) return NULL;
+
+	return get_form_info(ses, doc_view);
 }
